@@ -5,6 +5,7 @@ use strict;
 use base 'Exporter';
 our @EXPORT = qw(get_keyring);
 use Carp;
+use Passwd::Keyring::Auto::Chooser;
 
 =head1 NAME
 
@@ -12,11 +13,11 @@ Passwd::Keyring::Auto - interface to secure password storage(s)
 
 =head1 VERSION
 
-Version 0.2801
+Version 0.3001
 
 =cut
 
-our $VERSION = '0.2801';
+our $VERSION = '0.3001';
 
 =head1 SYNOPSIS
 
@@ -26,7 +27,8 @@ like Gnome Keyring, KDE Wallet, OSX/Keychain etc.
 
 While modules like Passwd::Keyring::Gnome handle specific backends,
 Passwd::Keyring::Auto tries to pick the best backend available,
-considering the current desktop environment.
+considering the current desktop environment, program options, and user
+configuration.
 
     use Passwd::Keyring::Auto;  # get_keyring
 
@@ -44,10 +46,10 @@ considering the current desktop environment.
     }
 
 If any secure backend is available, password is preserved
-for successive runs, and user need not be prompted.
+for successive runs, and user need not be prompted again.
 
-The choice can be impacted by some environment variables and/or
-additional parameters, see C<get_keyring> documentation for details.
+The choice can be impacted by configuration file, some environment
+variables and/or additional parameters, see L</BACKEND SELECTION CRITERIA>.
 
 One can skip this module and be explicit if he or she knows which
 keyring is to be used:
@@ -56,10 +58,6 @@ keyring is to be used:
     my $keyring = Passwd::Keyring::Gnome->new();
     # ... from there as above
 
-=head1 EXPORT
-
-get_keyring
-
 =head1 SUBROUTINES/METHODS
 
 =head2 get_keyring
@@ -68,6 +66,9 @@ get_keyring
 
     my $ring = get_keyring(app=>'MyApp', group=>'SyncPasswords');
 
+    my $ring = get_keyring(app=>'MyApp', group=>'Uploads',
+                           config=>"$ENV{HOME}/.passwd-keyring-business.cfg");
+
     my $ring = get_keyring(app=>'MyApp', group=>'Scrappers',
                            prefer=>['Gnome', 'PWSafe3'],
                            forbid=>['KDEWallet']);
@@ -79,15 +80,20 @@ get_keyring
                            %backend_specific_options);
 
 Returns the keyring object most appropriate for the current system
-(and matching specified criteria) and initiates it.
+(and matching specified criteria, and applying user configuration) and
+initiates it.
 
 The function inspects context the application runs in (operating
-system, presence of GUI sessions etc), decides which backends seem suitable
-and in what order of preference, then tries all suitable backends and returns
-first succesfully loaded and initialized.
+system, presence of GUI sessions etc), decides which backends seem
+suitable and in what order of preference, then tries all suitable
+backends and returns first succesfully loaded and initialized (or
+croaks if there is none). See L</BACKEND SELECTION CRITERIA> for
+info about criteria used.
 
-All parameters are optional, but it is recommended to set app and
-group:
+All parameters are optional, but it is strongly recommended to set
+C<app> and C<group>.
+
+General parameters:
 
 =over 4
 
@@ -97,7 +103,8 @@ Symbolic application name, which - depending on backend - may appear
 in interactive prompts (like dialog box "Application APP-NAME wants
 to access secure data..." popped up by KDE Wallet) and may be
 preserved as comment ("Created by ...") in secure storage (so may be
-seen in GUI password management apps like seahorse).
+seen in GUI password management apps like seahorse). Also, if config
+file is in use, it can override some settings on per-application basis.
 
 =item group => 'PasswordFolder'
 
@@ -108,6 +115,17 @@ apps/scripts will share passwords if they use the same group name, but
 will use different and unrelated passwords if they specify different
 group.
 
+=item config => "/some/where/passwd_keyring.cfg"
+
+Config file location.
+
+=back
+
+Parameters impacting backend selection (usually not recommended as
+they limit user choice, but hardcode choices if you like):
+
+=over 4
+
 =item force => 'Backend'
 
 Try only given backend and nothing else. Expects short backend name.
@@ -128,31 +146,115 @@ own preference.
 
 Never use specified backend(s).
 
-For example C<forbid=>['Gnome', 'KDEWallet']> will disable attempts to
-use GUI keyrings even if we run on Linux and have Gnome or KDE session
-active.
+For example C<forbid=>['Gnome', 'KDEWallet']> will cause method not to
+consider those GUI keyrings even if we run on Linux and have Gnome or
+KDE session active.
+
+=back
+
+Backend-specific parameters:
+
+=over 4
 
 =item other parameters
 
 All other parameters are passed as such to actual keyring backend.
 To check whether/which may be used, consult backends documentation.
-In general backends ignore params they do not know.
+Backends ignore params they do not understand, so some superset
+of possibly useful params is OK.
+
+It is recommended to use configuration file instead.
 
 =back
 
-The function should not in normal circumstances fail (there always is
-L<Passwd::Keyring::Memory> to be used if everything else fails), but
-it may croak if some keyring is enforced or if Memory is forbidden or
-uninstalled.
+The function in it's simplest form should not fail (it falls back to
+L<Passwd::Keyring::Memory> if everything else fails), but it may croak
+if some keyring is enforced or if Memory is forbidden or uninstalled.
 
 =head1 KEYRING METHODS
 
 See L<Passwd::Keyring::Auto::KeyringAPI> for operations available on
 keyring objects.
 
+=head1 CONFIGURATION FILE
+
+The recommended way to impact backend selection on per-system (and
+user) basis is to use configuration file, which let the user set
+default keyring selection rules, and per-application overrides.
+
+It's initial version can be created by L<passwd_keyring> script:
+
+    passwd_keyring configure
+
+and edited afterwards.
+
+See L</BACKEND SELECTION CRITERIA> for info how configuration settings
+relate to other backend selection methods.
+
+=head2 CONFIGURATION FILE LOCATION
+
+By default, config file is looked in C<~/.passwd-keyring.cfg> on
+Linux/Unix and C<~/Local Settings/Application Data/.passwd-keyring.cfg> 
+on Windows (more exactly: C<.passwd-keyring.cfg> in directory reported by
+C<my_data> function from L<File::HomeDir>).
+
+Environment variable C<PASSWD_KEYRING_AUTO_CONFIG> can be used to
+override this setting (and should contain path of the configuration
+file). Also, C<config> parameter can be used in C<get_keyring> method
+(and takes precedence even over env variable).
+
+Note that while it is OK not to have config file at all, but it is an
+error (and causes exception) to have non-existing or inaccessible file
+pointed by parameter or environment variable.
+
+=head2 CONFIGURATION FILE SYNTAX
+
+Example:
+
+    ; Default settings
+    prefer=KDEWallet PWSafe3 Memory
+    forbid=Gnome
+    PWSafe3.file=~/passwd-keyring.pwsafe3
+
+    ; Overrides for app named WebScrapers
+    [WebScrapers]
+    force=Gnome
+
+    ; Overrides for app named XYZTests
+    [XYZTests]
+    force=PWSafe3
+    PWSafe3.file=~/tests/xyz/passwd-keyring-tokens.pwsafe3
+
+C<prefer>, C<forbid> and C<force> define appropriate steering values,
+as documented in L<Passwd::Keyring::Auto>. Space is used to separate
+multiple values.
+
+C<;> can be used to start line comments.
+
 =head1 ENVIRONMENT VARIABLES
 
-The following environment variables can be used to impact the module behaviour:
+The following environment variables can be used to impact the module
+behaviour.
+
+General configuration variables:
+
+=over 4
+
+=item C<PASSWD_KEYRING_AUTO_CONFIG>
+
+Defines location of the config file.
+
+=item C<PASSWD_KEYRING_AUTO_DEBUG>
+
+Log on stderr details about tried and selected backends (and errors
+faced while they are tried).
+
+=back
+
+Backend-selection variables (see L</BACKEND SELECTION CRITERIA> for
+info how they relate to other methods and note that using
+configuration file is usually recommended over setting those
+variables):
 
 =over 4
 
@@ -183,161 +285,65 @@ param.
 
 =back
 
-The following variable provides some additional logging:
+=head1 BACKEND SELECTION CRITERIA
+
+Backend selection is organized around 3 steering parameters: C<force>,
+C<forbid>, and C<prefer>.  For each of those, the value is looked in
+the following places (first found is returned):
 
 =over 4
 
-=item C<PASSWD_KEYRING_AUTO_DEBUG>
+=item hardcoded value (C<get_keyring> param),
 
-Log on stderr details about tried and selected backends (and errors
-faced while they are tried).
+=item environment variable (C<PASSWD_KEYRING_AUTO_...>)
+
+=item configuration file per-application setting 
+
+=item configuration file default setting
+
+=item library default
 
 =back
 
-=head1 BACKEND SELECTION AND PREFERENCE CRITERIA
+Each param is calculated separately, so one can have C<prefer>
+initialized from hardcoded value, C<forbid> taken from the config file
+and C<force> defined by C<PASSWD_KEYRING_AUTO_FORCE> environment
+variable. This may sometimes be confusing so use sparingly (and limit
+to config file unless you really have reason to do otherwise).
 
-By default (no C<force>, C<prefer> or C<forbid> params, no environment
-variables) the following criteria are used (note that those may change
-without warning and are described here just for illustration):
+Once calculated, those params are used in the following way:
 
 =over 4
 
-=item Linux/Unix
+=item if C<force> is set, this is just used and remaining params are ignored - module tries to load this backend and either returns it, or (if it failed) raises an exception;
 
-Passwd::Keyring::Gnome and Passwd::Keyring::KDEWallet are tried first
-(if Gnome session is detected, Gnome version is first, under KDE and
-in unclear context KDEWallet takes preference). If both fail, emergency
-Passwd::Keyring::Memory is returned.
+=item elsewhere, all known backends are enumerated, and filtered by C<forbid> (so only those not forbidden remain)
 
-=item Mac OS/X
+=item the remaining list is sorted according to position on C<prefer> 
 
-Passwd::Keyring::OSXKeychain is tried, if it does not work, 
-Passwd::Keyring::Memory is returned.
+=item those modules are tried in order, first which succesfully loaded and initialized is returned
 
-=item Windows
-
-Currently Passwd::Keyring::Memory is always returned (this is to
-change once Windows Vault backend is written).
+=item if nothing was found, module raises exception.
 
 =back
 
-Note: some backends are not considered unless asked for (for example
-L<Passwd::Keyring::PWSafe3> is not currently considered by default
-algorithm).
+The following library defaults are used:
 
+=over 4
+
+=item there is no default for C<force>;
+
+=item C<forbid> is calculated according to the operating system (so L<Passwd::Keyring::OSXKeychain> is forbidden everywhere except Mac OS/X, L<Passwd::Keyring::Gnome> is forbidden on Windows and Mac, etc);
+
+=item C<prefer> is calculated according to operating system and detected session characteristics (so, if Gnome or Ubuntu session is detected, L<Passwd::Keyring::Gnome> is preferred, and if we have KDE, we prefer L<Passwd::Keyring:KDEWallet>, etc).
+
+=back
 
 =cut
 
 sub get_keyring {
-    my %options = @_;
-
-    my $debug = $ENV{PASSWD_KEYRING_AUTO_DEBUG} ? 1 : 0;
-
-    my $force = $options{force} || $ENV{PASSWD_KEYRING_AUTO_FORCE} || '';;
-    my $forbid = $options{forbid} || [ split(/\s+/, $ENV{PASSWD_KEYRING_AUTO_FORBID} || '') ];
-    my $prefer = $options{prefer} || [ split(/\s+/, $ENV{PASSWD_KEYRING_AUTO_PREFER} || '') ];
-    delete $options{forbid}; delete $options{prefer}; delete $options{force};
-
-    unless(ref($forbid)) {
-        $forbid = [$forbid];
-    }
-    unless(ref($prefer)) {
-        $prefer = [$prefer];
-    }
-
-    #################################################################
-    # Fast path for force
-    #################################################################
-
-    if($force) {
-        my $keyring = _try_backend($force, $debug, %options);
-        return $keyring if $keyring;
-        croak "Can not load enforced keyring $force";
-    }
-
-    #################################################################
-    # Selection and scoring of possible options.
-    #################################################################
-
-    # Note: we prefer to check possibly wrong module than to miss some.
-
-    my %candidates =(  # name → score, score > 0 means possible
-        'Gnome' => 0,
-        'KDEWallet' => 0,
-        'OSXKeychain' => 0,
-        'Memory' => 1,
-        );
-
-    # Scoring: +1000 for preferred, +100 for session-related, +10 for
-    # sensible, +1 for possible
-
-    if($^O eq 'darwin') {
-        $candidates{'OSXKeychain'} += 100;
-    }
-
-    if( $ENV{DISPLAY} || $ENV{DESKTOP_SESSION} ) {
-        $candidates{'KDEWallet'} += 11; # To give it some boost, more portable
-        $candidates{'Gnome'} += 10;
-    }
-
-    if($ENV{GNOME_KEYRING_CONTROL}) {
-        $candidates{'Gnome'} += 100;
-    }
-
-    if($ENV{DBUS_SESSION_BUS_ADDRESS}) {
-        $candidates{'KDEWallet'} += 10;
-    }
-
-    $candidates{$_} += 1000 foreach (@$prefer);
-    delete $candidates{$_} foreach (@$forbid);
-
-    my @attempts = grep { $candidates{$_} > 0 } keys %candidates;
-
-    @attempts = sort { ($candidates{$b} <=> $candidates{$a})
-                       ||
-                       ($a cmp $b)
-                   } @attempts;
-
-    if($debug) {
-        print STDERR "[Passwd::Keyring::Auto] Selected candidates(score): ",
-          join(", ", map { "$_($candidates{$_})" } @attempts), "\n";
-    }
-
-    foreach my $keyring_name (@attempts) {
-        my $keyring = _try_backend($keyring_name, $debug, %options);
-        return $keyring if $keyring;
-    }
-
-    croak "Could not load any keyring backend (attempted: " . join(", ", @attempts) . ")";
-}
-
-# Loads module of given name or returns undef if it does not work
-sub _try_backend {
-    my ($backend_name, $debug, @options) = @_;
-
-    # Sanity check
-    unless($backend_name =~ /^[A-Za-z][A-Za-z0-9_]*$/) {
-        if($debug) {
-            print STDERR "[Passwd::Keyring::Auto] Ignoring illegal backend name: $backend_name\n";
-        }
-        return undef;
-    }
-
-    my $keyring;
-    my $require = "Passwd/Keyring/$backend_name.pm";
-    my $module = "Passwd::Keyring::$backend_name";
-    eval {
-        require $require;
-        $keyring = $module->new(@options);
-    };
-    if($debug) {
-        unless($@) {
-            print STDERR "[Passwd::Keyring::Auto] Succesfully initiated $module, returning it\n";
-        } else {
-            print STDERR "[Passwd::Keyring::Auto] Attempt to use $module failed, error: $@\n";
-        }
-    }
-    return $keyring;
+    my $chooser = Passwd::Keyring::Auto::Chooser->new(@_);
+    return $chooser->get_keyring();
 }
 
 =head1 FURTHER INFORMATION
