@@ -1,2 +1,4 @@
 package Passwd::Keyring::Auto::KeyringAPI;
+use strict;
+use warnings;
 1;
